use crate::assets::asset_manager::AssetManager;

use crate::prelude::*;

use anyhow::Result;

pub trait AssetLoader {
    fn load_entity(&mut self, assets: AssetHandler<'_>, entity_file: &str) -> Result<EntityObject>;
}

pub struct AssetHandler<'a> {
    world: &'a mut World,
    asset_manager: &'a mut AssetManager,
    resource_base_path: String,
}

impl<'a> AssetHandler<'a> {
    pub fn create(world: &'a mut World) -> Self {
        let asset_manager = world.resources.get_mut_unchecked::<AssetManager>();
        let engine = world.resources.get::<Engine>();

        Self {
            asset_manager,
            resource_base_path: engine.resource_base_path.clone(),
            world,
        }
    }

    pub fn update_file_cache(&mut self) -> Result<()> {
        self.asset_manager.update_files()
    }

    pub fn preload_assets<'b>(&self, files: impl IntoIterator<Item = &'b &'b str>) -> Result<()> {
        self.asset_manager.async_load_gltf_file(self.world, files)
    }

    pub fn create_cache<'b>(
        &mut self,
        names: impl IntoIterator<Item = &'b &'b str> + Clone,
    ) -> Result<AssetCache> {
        self.asset_manager.create_cache_with(self.world, names)
    }

    pub fn insert_cache(&self, cache: &AssetCache) -> Result<()> {
        self.asset_manager.insert_cached_assets(cache)
    }

    pub fn clear(&mut self) -> Result<()> {
        self.asset_manager.clear()
    }

    pub fn gltf_files(&self) -> Result<Vec<String>> {
        Ok(self.asset_manager.gltf_files().clone())
    }

    pub fn create_entity(&mut self, asset_file: &str) -> Result<EntityObject> {
        self.asset_manager.load_asset_file(self.world, asset_file)
    }

    pub fn create_entity_absolute(&mut self, asset_path: &str) -> Result<EntityObject> {
        self.asset_manager
            .load_asset_file_absolute(self.world, asset_path)
    }

    pub fn empty_entity(&mut self) -> EntityObject {
        self.world.new_entity()
    }

    pub fn resource_base_path(&self) -> &str {
        &self.resource_base_path
    }

    pub fn world(&mut self) -> &mut World {
        &mut self.world
    }
}
