use crate::prelude::*;

use anyhow::Result;
use context::ContextObject;

use std::any::Any;
use std::collections::HashMap;
use std::sync::{Arc, RwLock};

// use super::engine_object_data::{EngineObjectAccess, EngineObjectDataHandle};

#[derive(Debug)]
pub enum EngineEvent {
    MouseMotion(u32, u32),
    MouseButtonDown(MouseButton),
    MouseButtonUp(MouseButton),
    MouseWheel(i32, i32, MouseWheelDirection),

    KeyDown(Keycode),
    KeyUp(Keycode),

    ButtonDown(ControllerButton),
    ButtonUp(ControllerButton),
    ControllerAxis(ControllerAxis),

    ControllerAdded(Arc<RwLock<Controller>>),
    ControllerRemoved(Arc<RwLock<Controller>>),

    FileDrop(String),
}

// pub trait EngineObjectHelper: Any {
//     type Payload: Send + Sync;

//     fn access(&self) -> EngineObjectAccess<'_, Self::Payload>;
//     fn payload(&self) -> EngineObjectDataHandle<Self::Payload>;
//     fn ui(&self) -> &States;
//     fn ui_mut(&mut self) -> &mut States;

//     fn access_callback_mut<'a, F>(
//         &'a self,
//         f: F,
//     ) -> impl FnMut() -> Result<()> + Send + Sync + 'static
//     where
//         F: FnMut(EngineObjectAccess<'a, Self::Payload>) -> Result<()> + Send + Sync + 'static,
//     {
//         self.payload().access_callback_mut(f)
//     }

//     fn access_callback<F>(&self, f: F) -> impl Fn() -> Result<()> + Send + Sync + 'static
//     where
//         F: Fn(EngineObjectAccess<'_, Self::Payload>) -> Result<()> + Send + Sync + 'static,
//     {
//         self.payload().access_callback(f)
//     }

//     fn handle_callback_mut<F>(&self, mut f: F) -> impl FnMut() -> Result<()> + Send + Sync + 'static
//     where
//         F: FnMut(
//                 EngineObjectDataHandle<Self::Payload>,
//                 EngineObjectAccess<'_, Self::Payload>,
//             ) -> Result<()>
//             + Send
//             + Sync
//             + 'static,
//     {
//         let handle = self.payload();

//         move || {
//             let data = handle.as_data();
//             let access = data.access();

//             f(handle.copy(), access)
//         }
//     }

//     fn handle_callback<F>(&self, f: F) -> impl Fn() -> Result<()> + Send + Sync + 'static
//     where
//         F: Fn(
//                 EngineObjectDataHandle<Self::Payload>,
//                 EngineObjectAccess<'_, Self::Payload>,
//             ) -> Result<()>
//             + Send
//             + Sync
//             + 'static,
//     {
//         let handle = self.payload();

//         move || {
//             let data = handle.as_data();
//             let access = data.access();

//             f(handle.copy(), access)
//         }
//     }
// }

pub trait EngineObject: Any {
    fn name(&self) -> &str;

    fn update(&mut self) -> Result<()>;
    fn event(&mut self, event: EngineEvent) -> Result<()>;
}

pub struct Input {
    pub direction_mapping: HashMap<Keycode, GuiDirection>,
}

pub struct ContextObjectImpl<E>
where
    E: EngineObject + Send + Sync,
{
    engine_object: E,

    context: Arc<Context>,
    input: Arc<RwLock<Input>>,
    gui_handler: Arc<GuiHandler>,

    ctrl_pressed: bool,
}

impl<E> ContextObjectImpl<E>
where
    E: EngineObject + Send + Sync,
{
    pub fn new(
        context: Arc<Context>,
        engine_object: E,
        input: Arc<RwLock<Input>>,
        gui_handler: Arc<GuiHandler>,
    ) -> Result<Self> {
        Ok(ContextObjectImpl {
            context,
            engine_object,
            input,
            gui_handler,
            ctrl_pressed: false,
        })
    }
}

impl<E> ContextObject for ContextObjectImpl<E>
where
    E: EngineObject + Send + Sync,
{
    fn name(&self) -> &str {
        self.engine_object.name()
    }

    fn update(&mut self) -> anyhow::Result<()> {
        self.gui_handler.process_callbacks()?;
        self.engine_object.update()
    }

    fn event(&mut self, event: Event) -> anyhow::Result<()> {
        match event {
            Event::MouseMotion(x, y) => {
                self.gui_handler.set_mouse_pos(x, y)?;
                self.engine_object.event(EngineEvent::MouseMotion(x, y))?;
            }
            Event::MouseButtonDown(mouse_button) => {
                if !self.gui_handler.mouse_down(mouse_button)? {
                    self.engine_object
                        .event(EngineEvent::MouseButtonDown(mouse_button))?;
                }
            }
            Event::MouseButtonUp(mouse_button) => {
                if !self.gui_handler.mouse_up(mouse_button)? {
                    self.engine_object
                        .event(EngineEvent::MouseButtonUp(mouse_button))?;
                }
            }
            Event::MouseWheel(x, y, direction) => self
                .engine_object
                .event(EngineEvent::MouseWheel(x, y, direction))?,
            Event::KeyDown(keycode) => {
                self.key_down_event(keycode)?;
            }
            Event::KeyUp(keycode) => {
                self.key_up_event(keycode)?;
            }
            Event::TextInput(text) => {
                self.text_input(text)?;
            }
            Event::ControllerButtonDown(button) => {
                self.button_down_event(button)?;
            }
            Event::ControllerButtonUp(button) => {
                self.button_up_event(button)?;
            }
            Event::ControllerAxis(controller) => {
                let controller = controller.read().unwrap();

                if !self.gui_handler.check_navigatable()? {
                    self.axis_event(&controller)?
                } else {
                    self.gui_handler.update_selection(controller.direction())?;
                }
            }
            Event::ControllerAdded(controller) => self.controller_added(controller)?,
            Event::ControllerRemoved(controller) => self.controller_removed(controller)?,
            Event::FileDrop(filename) => {
                self.engine_object.event(EngineEvent::FileDrop(filename))?
            }
        }

        Ok(())
    }
}

impl<E> ContextObjectImpl<E>
where
    E: EngineObject + Send + Sync,
{
    #[inline]
    fn controller_added(&mut self, controller: Arc<RwLock<Controller>>) -> Result<()> {
        self.engine_object
            .event(EngineEvent::ControllerAdded(controller))?;

        Ok(())
    }

    #[inline]
    fn controller_removed(&mut self, controller: Arc<RwLock<Controller>>) -> Result<()> {
        self.engine_object
            .event(EngineEvent::ControllerRemoved(controller))?;

        Ok(())
    }

    #[inline]
    fn key_up_event(&mut self, keycode: Keycode) -> Result<()> {
        let input = self.input.read().unwrap();

        if input.direction_mapping.get(&keycode).is_some()
            && self.gui_handler.update_selection(GuiDirection::None)?
        {
            return Ok(());
        }

        match keycode {
            Keycode::RCtrl | Keycode::LCtrl => {
                self.ctrl_pressed = false;
            }

            _ => (),
        }

        self.engine_object.event(EngineEvent::KeyUp(keycode))?;

        Ok(())
    }

    #[inline]
    fn key_down_event(&mut self, keycode: Keycode) -> Result<()> {
        let input = self.input.read().unwrap();

        if let Some(direction) = input.direction_mapping.get(&keycode) {
            if self.gui_handler.update_selection(*direction)? {
                return Ok(());
            }
        }

        match keycode {
            Keycode::Backspace => {
                if self.gui_handler.remove_char()? {
                    return Ok(());
                }
            }
            Keycode::Return => {
                if self.gui_handler.accept_selection()? {
                    return Ok(());
                }
            }
            Keycode::Escape => {
                if self.gui_handler.decline_topgui()? {
                    return Ok(());
                }
            }
            Keycode::RCtrl | Keycode::LCtrl => {
                self.ctrl_pressed = true;
            }
            Keycode::V => {
                if let Some(writeable) = self.gui_handler.writeable()? {
                    if let Some(content) = self.context.window_config().clipboard_content()? {
                        writeable.set_text(content)?;
                    }
                }
            }

            _ => (),
        }

        self.engine_object.event(EngineEvent::KeyDown(keycode))?;

        Ok(())
    }

    #[inline]
    fn button_up_event(&mut self, button: ControllerButton) -> Result<()> {
        self.engine_object.event(EngineEvent::ButtonUp(button))?;

        Ok(())
    }

    #[inline]
    fn button_down_event(&mut self, button: ControllerButton) -> Result<()> {
        if self.gui_handler.check_navigatable()? {
            self.check_button_down(button)?;
        } else {
            self.engine_object.event(EngineEvent::ButtonDown(button))?;
        }

        Ok(())
    }

    #[inline]
    fn axis_event(&mut self, controller: &Controller) -> Result<()> {
        self.engine_object
            .event(EngineEvent::ControllerAxis(controller.controller_axis()))?;

        Ok(())
    }

    #[inline]
    fn check_button_down(&mut self, button: ControllerButton) -> Result<()> {
        match button {
            ControllerButton::A => {
                if self.gui_handler.accept_selection()? {
                    return Ok(());
                }
            }

            ControllerButton::B => {
                if self.gui_handler.decline_topgui()? {
                    return Ok(());
                }
            }

            ControllerButton::RightButton => {
                if self.gui_handler.next_tab_topgui(false)? {
                    return Ok(());
                }
            }

            ControllerButton::LeftButton => {
                if self.gui_handler.previous_tab_topgui(false)? {
                    return Ok(());
                }
            }

            ControllerButton::RightTrigger => {
                if self.gui_handler.next_tab_topgui(true)? {
                    return Ok(());
                }
            }

            ControllerButton::LeftTrigger => {
                if self.gui_handler.previous_tab_topgui(true)? {
                    return Ok(());
                }
            }

            ControllerButton::DPadDown => {
                let selection_res = self.gui_handler.update_selection(GuiDirection::Down)?;
                self.gui_handler.update_selection(GuiDirection::None)?;

                if selection_res {
                    return Ok(());
                }
            }

            ControllerButton::DPadUp => {
                let selection_res = self.gui_handler.update_selection(GuiDirection::Up)?;
                self.gui_handler.update_selection(GuiDirection::None)?;

                if selection_res {
                    return Ok(());
                }
            }

            ControllerButton::DPadRight => {
                let selection_res = self.gui_handler.update_selection(GuiDirection::Right)?;
                self.gui_handler.update_selection(GuiDirection::None)?;

                if selection_res {
                    return Ok(());
                }
            }

            ControllerButton::DPadLeft => {
                let selection_res = self.gui_handler.update_selection(GuiDirection::Left)?;
                self.gui_handler.update_selection(GuiDirection::None)?;

                if selection_res {
                    return Ok(());
                }
            }

            _ => (),
        }

        if !self.gui_handler.accept_custom_selection(button)? {
            self.engine_object.event(EngineEvent::ButtonDown(button))?;
        }

        Ok(())
    }

    #[inline]
    fn text_input(&self, text: String) -> Result<()> {
        if let Some(writeable) = self.gui_handler.writeable()? {
            for c in text.chars() {
                writeable.add_letter(c)?;
            }
        }

        Ok(())
    }
}
