use std::sync::Arc;
use std::time::Duration;

use crate::prelude::*;

use anyhow::Result;

pub struct Animation {
    animations: Arc<Vec<Arc<asset::GltfAnimation>>>,

    current_animation: Option<Arc<asset::GltfAnimation>>,

    animation_time: Duration,

    index: Option<u32>,
    start_time: Duration,
    repeat: bool,
}

impl Animation {
    pub(crate) fn new(
        // animation_map: HashMap<AnimationType, Vec<u32>>,
        animations: Arc<Vec<Arc<asset::GltfAnimation>>>,
        draw: &Draw,
    ) -> Result<Animation> {
        let me = Animation {
            // animation_map,
            animations,

            current_animation: None,
            // animation_type: None,
            animation_time: Default::default(),

            index: None,
            start_time: Duration::default(),
            repeat: false,
        };

        if !me.animations.is_empty() {
            me.reset(draw)?;
        }

        Ok(me)
    }

    pub fn animations(&self) -> &Arc<Vec<Arc<asset::GltfAnimation>>> {
        &self.animations
    }

    pub fn set_repeat(&mut self, repeat: bool) {
        self.repeat = repeat;
    }

    fn reset_animation_index(&mut self) {
        self.index = None;
    }

    /// # Safety
    ///
    /// Make sure `now` is `scene_time`, otherwise you will experience strange behaviour
    pub unsafe fn set_animation(
        &mut self,
        draw: &Draw,
        index: Option<u32>,
        now: Duration,
        repeat: bool,
    ) -> Result<()> {
        match index {
            Some(index) => {
                self.current_animation = Some(self.animations[index as usize].clone());

                self.index = Some(index);
                self.start_time = now;
                self.repeat = repeat;
            }
            None => {
                if self.current_animation.is_some() {
                    self.current_animation = None;

                    self.reset_animation_index();
                    self.reset(draw)?;
                }
            }
        }

        Ok(())
    }

    pub fn reset(&self, draw: &Draw) -> Result<()> {
        // all animation have access to the same skins,
        // therefore reset will be the same for all animations
        if let Some(animation) = self.animations.get(0) {
            // reset bone matrices
            for mesh in draw.iter() {
                if let Some(bone_buffer) = mesh.bone_buffer() {
                    animation.reset(mesh.node_index, &mut unsafe {
                        bone_buffer.map_complete()?.reinterpret()
                    })?;
                }
            }
        }

        Ok(())
    }

    // HOW TO ANIMATE:
    //
    // iterate through meshes (every mesh is linked to a node)
    // check if that node has a skin -> animate through that skin
    pub fn animate(&mut self, time: Duration, draw: &Draw) -> Result<()> {
        self.check_time(time);

        // check if there is an animation set
        if let Some(animation) = &self.current_animation {
            // iterate over every mesh, that has:
            //  - a bone buffer
            //  - and whose node can be found (should never fail!?)
            // and then animate with that info
            for mesh in draw.iter() {
                if let Some(bone_buffer) = mesh.bone_buffer() {
                    animation.animate(self.animation_time, mesh.node_index, &mut unsafe {
                        bone_buffer.map_complete()?.reinterpret()
                    })?;
                }
            }
        }

        Ok(())
    }

    fn check_time(&mut self, now: Duration) {
        if let Some(animation) = &self.current_animation {
            let animation_duration = animation.duration();

            if self.start_time > now {
                self.index = None;
                self.current_animation = None;

                return;
            }

            self.animation_time = now - self.start_time;

            while self.animation_time > animation_duration {
                if !self.repeat {
                    self.index = None;
                    self.current_animation = None;

                    break;
                }

                self.start_time += animation_duration;
                self.animation_time = now - self.start_time;
            }
        }
    }
}

impl Clone for Animation {
    fn clone(&self) -> Self {
        Self {
            // animation_map: self.animation_map.clone(),
            animations: self.animations.clone(),

            current_animation: None,
            // animation_type: None,
            animation_time: Default::default(),

            index: None,
            start_time: Duration::default(),
            repeat: false,
        }
    }
}

impl EntityComponent for Animation {
    fn name(&self) -> &str {
        Self::debug_name()
    }
}

impl ComponentDebug for Animation {
    fn debug_name() -> &'static str {
        "Animation"
    }
}
