use crate::{prelude::*, scene::rendering::rasterizer::lights::light_wrapper::LightHandler};

use anyhow::Result;
use utilities::prelude::cgmath::{vec2, vec3};

use super::super::{
    super::{shared::safecommandbuffer::SafeCommandBuffer, RasterizerDescriptorLayouts},
    pipelines::RasterizerPipelines,
    rasterizershader::RasterizerShader,
};

use std::sync::{Arc, Mutex};

#[derive(Default)]
pub struct FallbackInfo {
    pub spot_light_count: u32,
    pub directional_light_count: u32,
    pub point_light_count: u32,
}

pub struct DeferredHandler {
    device: Arc<Device>,
    queue: Arc<Mutex<Queue>>,

    valid: Vec<bool>,

    screen_quad: Arc<Buffer<PositionColorUV>>,
    deferred_commands: SafeCommandBuffer<PositionColorUV>,
}

impl DeferredHandler {
    pub fn new(
        device: &Arc<Device>,
        queue: &Arc<Mutex<Queue>>,
        image_count: usize,
        render_pass: &Arc<RenderPass>,
        fallback: Option<FallbackInfo>,
    ) -> Result<Self> {
        let deferred_commands =
            Self::create_commands(device, queue, image_count, render_pass, fallback)?;

        let screen_quad = Self::create_screen_quad_buffer(device)?;

        Ok(DeferredHandler {
            device: device.clone(),
            queue: queue.clone(),

            valid: vec![false; image_count],

            screen_quad,

            deferred_commands,
        })
    }

    pub fn invalidate(&mut self) {
        for valid in &mut self.valid {
            *valid = false;
        }

        self.deferred_commands.invalidate();
    }

    pub fn render(
        &mut self,
        buffer_recorder: &mut CommandBufferRecorder<'_>,
        deferred_descriptors: &[&Arc<DescriptorSet>],
        framebuffer: &Arc<Framebuffer>,
        current_index: usize,
    ) -> Result<()> {
        let deferred_command_buffer = if !self.valid[current_index] {
            self.valid[current_index] = true;

            // record raytracing commands
            self.record(current_index, deferred_descriptors, framebuffer)?
        } else {
            let ray_cb = match self.deferred_commands.current(current_index) {
                Some(cb) => cb,
                None => panic!("invalid raytracing command buffer requested!"),
            };

            ray_cb
        };

        buffer_recorder.execute_commands(&[deferred_command_buffer.as_ref()]);

        Ok(())
    }

    pub fn resize_fallback_shader(
        &mut self,
        fallback: FallbackInfo,
        render_pass: &Arc<RenderPass>,
    ) -> Result<()> {
        self.deferred_commands = Self::create_commands(
            &self.device,
            &self.queue,
            self.valid.len(),
            render_pass,
            Some(fallback),
        )?;

        Ok(())
    }

    fn create_commands(
        device: &Arc<Device>,
        queue: &Arc<Mutex<Queue>>,
        image_count: usize,
        render_pass: &Arc<RenderPass>,
        fallback: Option<FallbackInfo>,
    ) -> Result<SafeCommandBuffer<PositionColorUV>> {
        let deferred_layout =
            RasterizerDescriptorLayouts::deferred_descriptor_layout(device.clone())?;
        let light_info_layout =
            LightHandler::light_info_descriptor_layout(device, fallback.is_some(), 32)?;
        let light_layout = LightHandler::light_descriptor_layout(device, fallback.is_some(), 32)?;

        let (vertex_shader, fragment_shader) = match fallback {
            Some(fallback) => RasterizerShader::deferred_fallback(
                device,
                0.7,
                fallback.spot_light_count,
                fallback.directional_light_count,
                fallback.point_light_count,
            )?,
            None => RasterizerShader::deferred(device, 0.7)?,
        };

        let pipeline_layout = PipelineLayout::builder()
            .add_descriptor_set_layout(&deferred_layout)
            .add_descriptor_set_layout(&light_info_layout)
            .add_descriptor_set_layout(&light_layout)
            .add_descriptor_set_layout(&light_info_layout)
            .add_descriptor_set_layout(&light_layout)
            .add_descriptor_set_layout(&light_info_layout)
            .add_descriptor_set_layout(&light_layout)
            .build(device.clone())?;

        let pipeline = RasterizerPipelines::create_default_pipeline::<PositionColorUV>(
            device.clone(),
            &pipeline_layout,
            vertex_shader,
            fragment_shader,
            render_pass,
            &[true],
            1,
        )?;

        SafeCommandBuffer::<PositionColorUV>::new_rasterize(
            device,
            queue,
            image_count,
            render_pass,
            1,
            pipeline,
        )
    }
}

impl DeferredHandler {
    fn record(
        &mut self,
        current_index: usize,
        deferred_descriptors: &[&Arc<DescriptorSet>],
        framebuffer: &Arc<Framebuffer>,
    ) -> Result<&Arc<CommandBuffer>> {
        let render_area = VkExtent2D {
            width: framebuffer.width(),
            height: framebuffer.height(),
        };

        let viewport = [VkViewport {
            x: 0.0,
            y: 0.0,
            width: render_area.width as f32,
            height: render_area.height as f32,
            minDepth: 0.0,
            maxDepth: 1.0,
        }];

        let scissor = [VkRect2D {
            extent: render_area,
            offset: VkOffset2D { x: 0, y: 0 },
        }];

        let mut cb = self
            .deferred_commands
            .begin(Some(framebuffer), current_index)?;

        cb.set_viewport(&viewport);
        cb.set_scissor(&scissor);

        self.deferred_commands
            .render(&mut cb, deferred_descriptors, &self.screen_quad)?;

        self.deferred_commands.end(cb, current_index)
    }

    fn create_screen_quad_buffer(device: &Arc<Device>) -> Result<Arc<Buffer<PositionColorUV>>> {
        let quad = [
            PositionColorUV::new(vec3(-1.0, -1.0, 0.0), vec2(0.0, 0.0)),
            PositionColorUV::new(vec3(1.0, -1.0, 0.0), vec2(1.0, 0.0)),
            PositionColorUV::new(vec3(1.0, 1.0, 0.0), vec2(1.0, 1.0)),
            PositionColorUV::new(vec3(1.0, 1.0, 0.0), vec2(1.0, 1.0)),
            PositionColorUV::new(vec3(-1.0, 1.0, 0.0), vec2(0.0, 1.0)),
            PositionColorUV::new(vec3(-1.0, -1.0, 0.0), vec2(0.0, 0.0)),
        ];

        Buffer::builder()
            .set_usage(VK_BUFFER_USAGE_VERTEX_BUFFER_BIT)
            .set_memory_usage(MemoryUsage::CpuOnly)
            .set_data(&quad)
            .build(device.clone())
    }
}
