use crate::game::{content::prelude::*, game::GameHandle};
use crate::Result;

use engine::prelude::*;

use cgmath::{vec2, InnerSpace, Vector2, Vector3};
use entity_manager::*;
use rpg_components::ability_type::AbilityType;
use rpg_components::components::ability_slots::AbilitySlots;
use rpg_components::components::character_status::CharacterStatus;
use rpg_components::damage_type::DamageType;

// space between user and ai hitbox
const SAFETY_RADIUS: f32 = 0.3;

const ZERO_MOVEMENT: Vector2<f32> = vec2(0.0, 0.0);

pub struct SimpleAI {
    pub aggro_radius: f32,
    pub deaggro_radius: f32,

    pub owner: Entity,

    pub target: Option<Entity>,
}

impl SimpleAI {
    pub fn new(aggro_radius: f32, deaggro_radius: f32, owner: Entity) -> Self {
        Self {
            aggro_radius,
            deaggro_radius,

            owner,

            target: None,
        }
    }

    pub fn update(
        &mut self,
        animation: &mut Animation,
        draw: &mut Draw,
        animation_info: &mut AnimationInfo,
        character_status: &mut CharacterStatus,
        movement: &mut Movement,
        location: &Location,
        abilities: &mut AbilitySlots,
        hitbox: &mut HitBox,
        scene_contents: &mut SceneContents<'_>,
        game: &GameHandle,
        user_entity: Entity,
    ) -> Result<()> {
        match self.check_target(location, hitbox, scene_contents, user_entity)? {
            Some((direction, distance_to_target, min_distance)) => {
                let mut can_move = true;

                for ability_opt in abilities.iter_mut() {
                    if let Some(book) = ability_opt {
                        // do nothing with this ability when user is out of range
                        if distance_to_target > book.ability().data().settings.parameter.distance {
                            continue;
                        }

                        let now = scene_contents.now();

                        // check if ability can be cast
                        if !animation_info.is_locked()
                            && book.validate_use(now, character_status, location)?
                        {
                            {
                                // TODO: further separation of animation types (bows, ...)
                                let animation_type =
                                    match book.ability().data().settings.parameter.damage_type {
                                        DamageType::Physical => AnimationType::Attack,
                                        _ => AnimationType::Cast,
                                    };

                                animation_info.set_animation(
                                    animation,
                                    draw,
                                    Some(animation_type),
                                    now,
                                    true,
                                    false,
                                )?;
                            }

                            let owner = unsafe {
                                scene_contents.entities.entity_mut_unchecked(self.owner)?
                            };

                            let mut multi_mut = owner.multi_mut();

                            // actually cast the ability
                            println!("ai casts ability");

                            match book.ability().data().settings.parameter.ability_type {
                                AbilityType::Projectile => Projectile::execute(
                                    book.ability().data(),
                                    self.owner,
                                    &mut multi_mut,
                                    location.direction(),
                                    book,
                                    game,
                                    &mut scene_contents.entities,
                                    &mut scene_contents.events,
                                )?,
                                AbilityType::SelfCast => todo!(),
                            }

                            can_move = false;

                            break;
                        }
                    }
                }

                if can_move {
                    if distance_to_target > min_distance {
                        movement.set_direction(direction);
                    } else {
                        movement.set_direction(ZERO_MOVEMENT);
                    }
                } else {
                    movement.set_direction(ZERO_MOVEMENT);
                }
            }

            None => {
                movement.set_direction(ZERO_MOVEMENT);
            }
        }

        Ok(())
    }

    #[inline]
    fn check_target(
        &mut self,
        location: &Location,
        hitbox: &HitBox,
        scene: &impl SceneEntities,
        user_entity: Entity,
    ) -> Result<Option<(Vector2<f32>, f32, f32)>> {
        let my_position = location.position();

        match &self.target {
            Some(target) => {
                let distance = match scene.entity(*target).ok() {
                    Some(object) => {
                        let target_position = object.get_component::<Location>()?.position();

                        (target_position - my_position).magnitude()
                    }

                    None => 1_000_000.0,
                };

                if distance > self.deaggro_radius {
                    self.target = None;
                    self.find_target(my_position, scene, user_entity)?;
                }
            }
            None => {
                self.find_target(my_position, scene, user_entity)?;
            }
        }

        self.target
            .as_ref()
            .map(|target| {
                let (position, radius) = {
                    let object = scene.entity(*target)?;

                    (
                        object.get_component::<Location>()?.position(),
                        object.get_component::<HitBox>()?.radius(),
                    )
                };

                let direction = position - my_position;
                let magnitude = direction.magnitude();

                Ok((
                    direction.truncate().normalize(),
                    magnitude,
                    radius + hitbox.radius() + SAFETY_RADIUS,
                ))
            })
            .transpose()
    }

    // If there is no target set, choose the closest user
    #[inline]
    fn find_target(
        &mut self,
        my_position: Vector3<f32>,
        scene: &impl SceneEntities,
        user_entity: Entity,
    ) -> Result<()> {
        assert!(self.target.is_none());

        let mut closest = None;

        if let Some(user_position) = scene
            .entity(user_entity)
            .ok()
            .map(|user_object| user_object.get_component::<Location>().unwrap().position())
        {
            let distance = (user_position - my_position).magnitude();

            if distance <= self.aggro_radius {
                match &closest {
                    Some((_entity, current_distance)) => {
                        if distance < *current_distance {
                            closest = Some((user_entity, distance));
                        }
                    }
                    None => {
                        closest = Some((user_entity, distance));
                    }
                }
            }
        }

        if let Some((entity, _)) = closest {
            self.target = Some(entity);
        }

        Ok(())
    }
}

impl EntityComponent for SimpleAI {
    fn name(&self) -> &str {
        Self::debug_name()
    }
}

impl ComponentDebug for SimpleAI {
    fn debug_name() -> &'static str {
        "SimpleAI"
    }
}
