use std::{collections::HashSet, time::Duration};

use cgmath::{vec2, Deg, InnerSpace, Matrix2, Vector2};
use engine::prelude::*;
use entity_manager::*;
use rpg_components::{
    components::{abilityloader::AbilityLoader, statistics::Statistics},
    damage_type::DamageType,
};

use crate::*;

use self::game::content::abilities::{damage_and_experience, handle_npc_death};

pub struct AreaOfEffectArc {
    arc: Deg<f32>,
    radius: f32,

    base_damage: u32,
    damage_type: DamageType,
    owner: Entity,

    spawn_time: Duration,
    collider: HashSet<Entity>,
}

impl AreaOfEffectArc {
    pub fn new(
        arc: impl Into<Deg<f32>>,
        radius: f32,
        base_damage: u32,
        damage_type: DamageType,
        owner: Entity,
    ) -> Self {
        let arc = arc.into();

        #[cfg(debug_assertions)]
        {
            let arc_deg: Deg<f32> = arc;

            if arc_deg.0 < 0.0 {
                panic!("arc should not be below 0°");
            }

            if arc_deg.0 > 180.0 {
                panic!("arc should not be above 180°");
            }
        }

        Self {
            arc,
            radius,

            base_damage,
            damage_type,
            owner,

            spawn_time: Duration::default(),
            collider: HashSet::new(),
        }
    }

    pub fn arc(&self) -> Deg<f32> {
        self.arc
    }

    pub fn radius(&self) -> f32 {
        self.radius
    }

    pub fn check_life_time(&self, now: Duration) -> bool {
        now <= self.spawn_time + Duration::from_secs(1)
    }

    /// returns true if the collider was not present before
    pub fn check_collider(&mut self, collider: Entity) -> bool {
        self.collider.insert(collider)
    }

    pub fn check_collision(
        &self,
        my_location: &Location,
        my_direction: Vector2<f32>,
        collider_hit_box: &HitBox,
        collider_location: &Location,
    ) -> bool {
        let collider_pos = collider_location.position().xy();
        let my_pos = my_location.position().xy();

        if !collider_hit_box.check_collision(collider_pos, self.radius, my_pos) {
            return false;
        }

        if self.arc.0 < 175.0 {
            let left_border = Matrix2::from_angle(self.arc) * my_direction;
            let right_border = Matrix2::from_angle(-self.arc) * my_direction;

            let left_normal = vec2(-left_border.y, left_border.x);
            let right_normal = vec2(right_border.y, -right_border.x);

            let diff = my_pos - collider_pos;
            let norm_diff = -diff.normalize();

            let left_dot = left_normal.dot(norm_diff);
            let right_dot = right_normal.dot(norm_diff);

            if left_dot > 0.0 && right_dot > 0.0 {
                return false;
            }

            let left_distance = (diff - (diff.dot(left_border)) * left_border).magnitude();
            let right_distance = (diff - (diff.dot(right_border)) * right_border).magnitude();

            if (left_distance > collider_hit_box.radius() && left_dot > 0.0)
                || (right_distance > collider_hit_box.radius() && right_dot > 0.0)
            {
                return false;
            }
        }

        true
    }

    pub fn hit_event(
        &mut self,
        game_handle: &GameHandle,
        scene: &mut SceneContents<'_>,
        collider: Entity,
    ) -> Result<()> {
        let s = unsafe { remove_life_time_mut(scene) };

        let mut entity_multi_mut = scene.entities_multi_mut();

        // collider
        let Ok(collider_object) = entity_multi_mut.get(collider) else {
            return Ok(());
        };
        let mut collider_multi_mut = collider_object.multi_mut();

        // owner
        let Ok(owner_object) = entity_multi_mut.get(self.owner) else {
            return Ok(());
        };
        let mut owner_multi_mut = owner_object.multi_mut();

        // factions
        let collider_faction = Faction::get(&mut collider_multi_mut);
        let owner_faction = Faction::get(&mut owner_multi_mut).expect("owner is not in a faction!");

        if !owner_faction.check_collision(collider_faction) {
            return Ok(());
        }

        unsafe {
            owner_multi_mut.clear_all_usages();
        }

        let owner_stats = owner_multi_mut.get::<Statistics>()?;
        let damage = AbilityLoader::check_for_crit(self.base_damage, owner_stats);

        damage_and_experience(
            damage,
            self.damage_type,
            &mut collider_multi_mut,
            owner_multi_mut,
            s,
        )?;

        unsafe {
            collider_multi_mut.clear_all_usages();
        }

        handle_npc_death(&game_handle.upgrade(), s, collider, collider_multi_mut)?;

        Ok(())
    }
}

impl EntityComponent for AreaOfEffectArc {
    fn name(&self) -> &str {
        Self::debug_name()
    }

    fn enable(&mut self, scene: &mut Scene) -> Result<()> {
        self.spawn_time = scene.now();

        Ok(())
    }

    fn disable(&mut self, _scene: &mut Scene) -> Result<()> {
        Ok(())
    }
}

impl ComponentDebug for AreaOfEffectArc {
    fn debug_name() -> &'static str {
        "AreaOfEffectArc"
    }
}

// #[cfg(test)]
// mod test {
//     use cgmath::{vec2, vec3, Deg};
//     use engine::prelude::*;

//     use crate::{AreaOfEffectArc, HitBox, Location};

//     #[test]
//     fn verify_aoe_hitbox_collision() {
//         let mut left_location = Location::new();
//         left_location.set_position(vec3(35.93969, 14.690732, 0.0));

//         let aoe_arc = AreaOfEffectArc::new(Deg(15.0), 0.8, |_, _, _| Ok(()));

//         let left_direction = vec2(0.9995991, 0.02831539);

//         let mut right_position = Location::new();
//         right_position.set_position(vec3(36.608078, 14.694146, 0.0));

//         let hitbox = HitBox::new(0.25689012, 1.0);

//         assert!(aoe_arc.check_collision(&left_location, left_direction, &hitbox, &right_position));
//     }
// }
