use cgmath::Vector3;
use rpg_components::damage_type::DamageType;

use crate::*;

use std::{sync::Arc, time::Duration};

pub struct DamageEvent {
    pub damage: u32,
    pub damage_type: DamageType,
    pub position: Vector3<f32>,
}

pub struct DamageNumber {
    grid: Arc<Grid>,
    start: Duration,
}

impl DamageNumber {
    pub const LIFE_TIME: Duration = Duration::from_secs(2);
    pub const SPEED: f32 = 1.5;

    pub fn new(game_handle: &GameHandle, damage: u32, damage_type: DamageType) -> Result<Self> {
        Ok(Self {
            grid: Self::create_ui(game_handle, damage, damage_type)?,
            start: Duration::default(),
        })
    }

    fn create_ui(
        game_handle: &GameHandle,
        damage: u32,
        damage_type: DamageType,
    ) -> Result<Arc<Grid>> {
        let game = game_handle.upgrade();
        let gui_handler = game.engine().gui_handler();

        let grid = Grid::new(gui_handler.clone(), 1, 1, false)?;
        grid.disallow_position_scale()?;

        let label = Label::builder()
            .set_text(damage.to_string())
            .set_text_color(damage_type)
            .build(gui_handler.clone())?;

        grid.attach(label, 0, 0, 1, 1)?;
        grid.set_frame(0, 0, 120, 40, VerticalAlign::Top, HorizontalAlign::Left)?;

        Ok(grid)
    }

    pub fn update(&self, scene: &SceneContents<'_>, location: &mut Location) -> Result<bool> {
        let diff = scene.now() - self.start;

        if diff > Self::LIFE_TIME {
            return Ok(false);
        }

        let ratio = diff.as_secs_f32() / Self::LIFE_TIME.as_secs_f32();
        let mut position = location.position();
        position.z += Self::SPEED * ratio;

        let (x, y) = scene.world_to_screen_space(position)?;
        let (width, height) = self.grid.extents();
        self.grid
            .set_position(x - (width as i32 / 2), y - (height as i32 / 2))?;

        Ok(true)
    }
}

impl DamageNumber {
    pub fn setup_event(game_handle: &GameHandle, scene: &mut Scene) {
        scene.register_event::<DamageEvent>();

        scene.add_event_reader({
            let game_handle = game_handle.clone();

            move |scene, damage_event: &DamageEvent| {
                let mut entity = game_handle.upgrade().engine().assets().empty_entity();

                let location = Location::new_and_setup(&mut entity)?;
                location.set_position(damage_event.position);

                let damage_number =
                    DamageNumber::new(&game_handle, damage_event.damage, damage_event.damage_type)?;
                entity.insert_component(damage_number);

                scene.add_entity(entity)?;

                Ok(())
            }
        });
    }
}

impl EntityComponent for DamageNumber {
    fn name(&self) -> &str {
        Self::debug_name()
    }

    fn enable(&mut self, scene: &mut Scene) -> Result<()> {
        self.start = scene.now();
        self.grid.set_visibility(true)
    }

    fn disable(&mut self, _scene: &mut Scene) -> Result<()> {
        self.grid.set_visibility(false)
    }
}

impl ComponentDebug for DamageNumber {
    fn debug_name() -> &'static str {
        "DamageNumber"
    }
}
